# Copyright (c) 2014-2025 Sebastian Noack
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

import sys
import os
import socket
import pickle
import subprocess
import ctypes.util

from mypass import Error, ConnectionLost, SOCKET
from mypass.config import get_config


def _spawn_daemon():
    lib = ctypes.util.find_library('sqlcipher')
    if not lib:
        raise Error('SQLCipher library is not installed')
    ld_preload = '{} {}'.format(os.environ.get('LD_PRELOAD', ''), lib).lstrip()
    process = subprocess.Popen(
        [sys.executable, '-m', 'mypass.daemon'],
        env=dict(os.environ, LD_PRELOAD=ld_preload),
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        bufsize=0
    )
    return (process.stdout, process.stdin)


class Client:
    def __init__(self, shared=True):
        self.shared = shared
        self._pipe = None

        if shared:
            try:
                self._connect()
            except (FileNotFoundError, ConnectionRefusedError):
                pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def _connect(self):
        with socket.socket(socket.AF_UNIX, socket.SOCK_STREAM) as sock:
            sock.connect(SOCKET)
            self._pipe = (sock.makefile('rb', 0),
                          sock.makefile('wb', 0))

    @property
    def database_locked(self):
        return not self._pipe

    def unlock_database(self, passphrase):
        self._pipe = _spawn_daemon()
        self.call('init', passphrase, self.shared)

    def call(self, command, *args):
        reader, writer = self._pipe
        try:
            pickle.dump((command, args), writer)
            output = pickle.load(reader)
        except (BrokenPipeError, EOFError):
            raise ConnectionLost

        if isinstance(output, Error):
            raise output

        return output

    def close(self):
        if self._pipe:
            self._pipe[0].close()
            self._pipe[1].close()


def database_exists():
    try:
        return os.stat(get_config('database', 'path')).st_size > 0
    except FileNotFoundError:
        return False



# Copyright (c) 2014-2025 Sebastian Noack
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

import sys
import random
import string
import argparse
from getpass import getpass

try:
    import argcomplete
except ImportError:
    argcomplete = None

from mypass import Error, CredentialsDoNotExist, CredentialsAlreadytExist
from mypass.client import Client, database_exists
from mypass.config import get_config


def generate_password(length):
    rand = random.SystemRandom()

    chars = [
        rand.choice(string.ascii_lowercase),
        rand.choice(string.ascii_uppercase),
        rand.choice(string.digits),
        rand.choice(string.punctuation),
    ]

    while len(chars) < length:
        chars.append(chr(rand.randint(33, 126)))

    rand.shuffle(chars)
    return ''.join(chars)[:length]


def prompt_new_passphrase():
    passphrase = getpass('New passphrase: ')

    if passphrase != getpass('Verify passphrase: '):
        print("Password doesn't match!", file=sys.stderr)
        sys.exit(1)

    return passphrase


def complete_context(**kwargs):
    with Client() as client:
        if not client.database_locked:
            return client.call('get-contexts')
    return []


def complete_username(parsed_args, **kwargs):
    with Client() as client:
        if not client.database_locked:
            try:
                return [user for user, _ in client.call('get-credentials',
                                                        parsed_args.context)]
            except CredentialsDoNotExist:
                pass
    return []


class CLI:

    def __init__(self, args=None):
        try:
            self._parser = get_argument_parser()

            if argcomplete:
                argcomplete.autocomplete(self._parser, default_completer=None)

            self._args = self._parser.parse_args(args=args)

            with Client() as self._client:
                self._open_database()
                getattr(self, '_call_' + self._args.command)()
        except Error as e:
            print(e, file=sys.stderr)
            sys.exit(1)
        except KeyboardInterrupt:
            print(file=sys.stderr)
            sys.exit(1)

    def _open_database(self):
        if self._client.database_locked:
            if getattr(self._args, 'exit_if_db_locked', False):
                sys.exit(0)

            if database_exists():
                passphrase = getpass('Unlock database: ')
            else:
                if getattr(self._args, 'fail_if_db_does_not_exist', False):
                    print('Database does not exist', file=sys.stderr)
                    sys.exit(1)

                passphrase = prompt_new_passphrase()

            self._client.unlock_database(passphrase)

    def _check_override(self, *args):
        try:
            self._client.call(*args)
        except CredentialsAlreadytExist:
            if input('Credentials already exist. Do you want to '
                     'override them? [y/N] ')[:1].lower() != 'y':
                print('Aborted', file=sys.stderr)
                sys.exit(1)

            self._client.call(*(args + (True,)))

    def _call_get(self):
        credentials = self._client.call('get-credentials', self._args.context)

        if len(credentials) == 1 and credentials[0][0] == '':
            print(credentials[0][1])
            return

        username_width = max(len(token[0]) for token in credentials) + 2
        for username, password in credentials:
            print(username.ljust(username_width), end='')
            print(password)

    def _call_add(self, password=None):
        self._check_override('store-credentials',
                             self._args.context,
                             self._args.username,
                             password or self._args.password or getpass())

    def _call_new(self):
        length = self._args.length

        if length is None:
            length = get_config('password', 'length')

        min_length = 1
        max_length = 10000
        if not (min_length <= length <= max_length):
            self._parser.error('Length of password must be '
                               'between {} and {}, at least 16 '
                               'is recommended'.format(min_length, max_length))

        password = generate_password(length)
        self._call_add(password)
        print(password)

    def _call_remove(self):
        context = self._args.context
        username = self._args.username

        if username is not None:
            self._client.call('delete-credentials', context, username)
        else:
            self._client.call('delete-context', context)

    def _call_rename(self):
        old_context = self._args.context
        new_context = self._args.new_context
        old_username = self._args.username
        new_username = self._args.new_username

        if new_context is None:
            new_context = old_context

        if old_username is not None or new_username is not None:
            self._check_override('rename-credentials',
                                 old_context, old_username or '',
                                 new_context, new_username or '')
        else:
            self._check_override('rename-context', old_context, new_context)

    def _call_list(self):
        for context in self._client.call('get-contexts'):
            print(context)

    def _call_alias(self):
        self._check_override('add-context-alias', self._args.context, self._args.alias)

    def _call_changepw(self):
        self._client.call('change-passphrase', prompt_new_passphrase())

    def _call_lock(self):
        self._client.call('shutdown')


def get_argument_parser():
    parser = argparse.ArgumentParser(
        prog='mypass',
        description='A secure password manager with command line interface',
    )
    parser.man_short_description = parser.description

    subparsers = parser.add_subparsers(dest='command')
    subparsers.required = True

    subparser_get = subparsers.add_parser('get', help='Writes the requested password to stdout')
    subparser_get.add_argument('context').completer = complete_context
    subparser_get.set_defaults(fail_if_db_does_not_exist=True)

    subparser_add = subparsers.add_parser('add', help='Adds the given password to the database')
    subparser_add.add_argument('context').completer = complete_context
    subparser_add.add_argument('username', nargs='?', default='')
    subparser_add.add_argument('password', nargs='?')

    subparser_new = subparsers.add_parser('new', help='Generates a new password and adds it to the database')
    subparser_new.add_argument('context').completer = complete_context
    subparser_new.add_argument('username', nargs='?', default='')
    subparser_new.add_argument('--length', '-l', type=int)

    subparser_remove = subparsers.add_parser('remove', help='Removes a password from the database')
    subparser_remove.add_argument('context').completer = complete_context
    subparser_remove.add_argument('username', nargs='?').completer = complete_username
    subparser_remove.set_defaults(fail_if_db_does_not_exist=True)

    subparser_rename = subparsers.add_parser('rename', help='Changes the context and/or username of saved passwords')
    subparser_rename.add_argument('context').completer = complete_context
    subparser_rename.add_argument('username', nargs='?').completer = complete_username
    subparser_rename.add_argument('--new-context').completer = complete_context
    subparser_rename.add_argument('--new-username')
    subparser_rename.set_defaults(fail_if_db_does_not_exist=True)

    subparser_list = subparsers.add_parser('list', help='Writes the contexts of all passwords to stdout')
    subparser_list.set_defaults(fail_if_db_does_not_exist=True)

    subparser_alias = subparsers.add_parser('alias', help='Creates a new context that refers to the credentials of an existing context')
    subparser_alias.add_argument('context').completer = complete_context
    subparser_alias.add_argument('alias')
    subparser_alias.set_defaults(fail_if_db_does_not_exist=True)

    subparser_changepw = subparsers.add_parser('changepw', help='Changes the master passphrase')
    subparser_changepw.set_defaults(fail_if_db_does_not_exist=True)

    subparser_lock = subparsers.add_parser('lock', help='Closes the database and forgets the master passhrase')
    subparser_lock.set_defaults(exit_if_db_locked=True)

    return parser



# Copyright (c) 2014-2025 Sebastian Noack
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

import os
import json
import tempfile
import hashlib

try:
    try:
        import Cryptodome.Cipher.AES
        Crypto = Cryptodome
    except ImportError:
        import Crypto.Cipher.AES
except ImportError:
    Crypto = None

import sqlite3

from mypass import CredentialsAlreadytExist, DatabaseError
from mypass.db import Database

BLOCK_SIZE = 16


def update_from_legacy_db(filename, passphrase):
    with open(filename, 'rb') as file:
        salt = file.read(48)
        iv = file.read(BLOCK_SIZE)
        ciphertext = file.read()

    if not ciphertext or len(ciphertext) % BLOCK_SIZE:
        return False

    if not Crypto:
        raise DatabaseError(
            'Wrong passphrase or legacy database (PyCrypto or PyCryptodome '
            "isn't installed, cannot detect or migrate legacy databases)"
        )

    for iterations in [200000, 10000]:
        key = hashlib.pbkdf2_hmac('sha256', passphrase.encode('utf-8'),
                                  salt, iterations, 32)
        cipher = Crypto.Cipher.AES.new(key, Crypto.Cipher.AES.MODE_CBC, iv)

        try:
            data = json.loads(cipher.decrypt(ciphertext).decode('utf-8'))
        except ValueError:
            continue

        fd, tempfilename = tempfile.mkstemp('.db', dir=os.path.dirname(filename))
        os.close(fd)

        try:
            with Database(tempfilename, passphrase) as db:
                for context, credentials in data.items():
                    for username, password in credentials.items():
                        db.store_credentials(context, username, password)
        except BaseException as e:
            os.unlink(tempfilename)
            if isinstance(e, (AttributeError,
                              CredentialsAlreadytExist,
                              sqlite3.InterfaceError)):
                continue
            raise

        os.rename(tempfilename, filename)

        return True

    return False



# Copyright (c) 2014-2020 Sebastian Noack
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

import os
import configparser

from mypass import ConfigError


class _Path:
    def __init__(self, path):
        self.path = path


DATA_DIR = '~/.config/mypass'

SCHEMA = {
    'daemon': {
        'timeout': 30,
        'logfile': _Path(os.path.join(DATA_DIR, 'log')),
    },
    'database': {
        'path': _Path(os.path.join(DATA_DIR, 'db')),
    },
    'password': {
        'length': 16,
    },
}

_parser = None


def get_config(section, option):
    global _parser

    if not _parser:
        _parser = configparser.ConfigParser()
        try:
            try:
                _parser.read(os.path.join(os.path.expanduser(DATA_DIR), 'config.ini'))
            except UnicodeDecodeError:
                raise ConfigError('Unexpected encoding in config file')
            except configparser.ParsingError:
                raise ConfigError('Invalid syntax in config file')
        except:
            _parser = None
            raise

    default = SCHEMA[section][option]
    is_path = False
    if isinstance(default, _Path):
        default = default.path
        is_path = True

    if isinstance(default, bool):
        func = _parser.getboolean
    else:
        func = getattr(_parser, 'get' + type(default).__name__, _parser.get)

    try:
        value = func(section, option, fallback=default)
    except ValueError:
        raise ConfigError('Invalid value for option {!r} in section '
                          '{!r} in config file'.format(option, section))

    if is_path:
        value = os.path.expanduser(value)

    return value



import os
import json
import re
import subprocess
import time
from datetime import datetime, timezone
from collections import OrderedDict
from distutils import log
from distutils.errors import DistutilsSetupError

from setuptools import Command, setup
from setuptools.command.bdist_wheel import bdist_wheel
from setuptools.command.build_py import build_py
from setuptools.command.sdist import sdist


MANPAGE = os.path.join('man', 'mypass.1')


def require_manpages_factory(base_command):
    class RequireManpages(base_command):
        def run(self):
            if not os.path.isfile(MANPAGE):
                raise DistutilsSetupError(
                    '{} is missing; run "python3 setup.py build_manpages" first.'.format(MANPAGE)
                )
            super().run()

    return RequireManpages


class BuildPy(build_py):
    def _write_manifest(self, package_dir, vendor):
        manifest = OrderedDict()
        manifest['name'] = 'org.snoack.mypass'
        manifest['description'] = self.distribution.metadata.description
        # pipx/pip users rewrite this template in the README instructions,
        # dh-python installs the console script into /usr/bin.
        manifest['path'] = '/usr/bin/mypass'
        manifest['type'] = 'stdio'

        if vendor == 'mozilla':
            manifest['allowed_extensions'] = ['mypass@snoack.addons.mozilla.org']
        else:
            manifest['allowed_origins'] = ['chrome-extension://ddbeciaedkkgeiaellofogahfcolmkka/']

        dirname = os.path.join(package_dir, 'native-messaging-hosts', vendor)
        self.mkpath(dirname)

        outfile = os.path.join(dirname, manifest['name'] + '.json')
        log.info('Writing %s', outfile)

        with open(outfile, 'w') as file:
            json.dump(manifest, file, indent='  ')

    def run(self):
        super().run()

        package_dir = os.path.join(self.build_lib, 'mypass')
        self._write_manifest(package_dir, 'chrome')
        self._write_manifest(package_dir, 'mozilla')
        self.copy_tree('extension', os.path.join(package_dir, 'extension'))


class BuildManpages(Command):
    description = 'Generate manual pages from README.md and argparse metadata'
    user_options = [
        ('output=', 'O', 'output manpage path'),
        ('readme=', 'R', 'README markdown path'),
    ]

    def initialize_options(self):
        self.output = None
        self.readme = None
        self.argument_parser = None

    def finalize_options(self):
        if self.output is None:
            self.output = MANPAGE
        if self.readme is None:
            self.readme = 'README.md'

    def _extract_section(self, readme, title):
        match = re.search(
            r'^{}\n-+\n(?P<body>.*?)(?=^[^\n]+\n[-=]+\n|\Z)'.format(re.escape(title)),
            readme,
            re.MULTILINE | re.DOTALL,
        )
        if not match:
            raise ValueError('Could not find README section: {}'.format(title))

        return match.group('body').strip()

    def _get_argument_parser(self):
        from mypass.cli import get_argument_parser
        return get_argument_parser()

    def _build_manpage_markdown(self, argument_parser):
        with open(self.readme) as file:
            readme = file.read()

        parts = [
            '# NAME',
            '',
            '{} - {}'.format(argument_parser.prog, self.distribution.get_description()),
            '',
            '# SYNOPSIS',
            '',
            '```text',
            argument_parser.format_usage().strip().split(': ', 1)[-1],
            '```',
        ]

        for title in ['Usage', 'Configuration']:
            parts.extend([
                '',
                '# {}'.format(title.upper()),
                '',
                self._extract_section(readme, title),
            ])

        return '\n'.join(parts).strip() + '\n'

    def run(self):
        argument_parser = self._get_argument_parser()
        build_timestamp = int(os.environ.get('SOURCE_DATE_EPOCH') or time.time())
        build_date = datetime.fromtimestamp(build_timestamp, timezone.utc)
        self.mkpath(os.path.dirname(self.output))
        subprocess.run(
            [
                'pandoc',
                '--standalone',
                '--from=gfm',
                '--to=man',
                '--output', self.output,
                '--metadata=title:{}(1) {} {}'.format(
                    argument_parser.prog.upper(),
                    argument_parser.prog,
                    self.distribution.get_version()
                ),
                '--metadata=author:{} <{}>'.format(
                    self.distribution.get_author(),
                    self.distribution.get_author_email()
                ),
                '--metadata=date:{}'.format(build_date.strftime('%Y-%m-%d'))
            ],
            input=self._build_manpage_markdown(argument_parser),
            text=True,
            check=True
        )


with open(os.path.join(os.path.dirname(__file__), 'README.md')) as file:
    long_description = file.read()

setup(name='mypass',
      description='A secure password manager with command line interface',
      long_description=long_description,
      long_description_content_type='text/markdown',
      author='Sebastian Noack',
      author_email='sebstian.noack@gmail.com',
      url='https://github.com/snoack/mypass',
      license='GPL-3.0-or-later',
      version='2.2',
      packages=['mypass'],
      scripts=['bin/mypass'],
      data_files=[('share/man/man1', [MANPAGE])],
      package_data={'mypass': ['extension/*', 'native-messaging-hosts/*/*.json']},
      install_requires=['pycryptodome', 'argcomplete'],
      python_requires='>=3.4',
      cmdclass={
          'build_py': BuildPy,
          'build_manpages': BuildManpages,
          'sdist': require_manpages_factory(sdist),
          'bdist_wheel': require_manpages_factory(bdist_wheel),
      },
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Intended Audience :: End Users/Desktop',
          'Intended Audience :: Information Technology',
          'Operating System :: POSIX',
          'Programming Language :: Python',
          'Programming Language :: Python :: 3',
          'Topic :: Security',
          'Topic :: Utilities',
      ])



# Copyright (c) 2014-2025 Sebastian Noack
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

from getpass import getuser

SOCKET = '/tmp/mypass-{}.sock'.format(getuser())


class Error(Exception):
    pass


class ConfigError(Error):
    pass


class DatabaseError(Error):
    pass


class CredentialsDoNotExist(Error):

    def __str__(self):
        return 'Credentials do not exist'


class CredentialsAlreadytExist(Error):

    def __str__(self):
        return 'Credentials already exist'


class ConnectionLost(Error):

    def __str__(self):
        return 'Connection lost'



# Copyright (c) 2014-2020 Sebastian Noack
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

import os

import sqlite3

from mypass import DatabaseError, CredentialsDoNotExist, CredentialsAlreadytExist


def makedir_wrapper(func, filename, exc=FileNotFoundError):
    try:
        return func(filename)
    except exc:
        os.makedirs(os.path.dirname(filename), exist_ok=True)
        return func(filename)


def _execute_pragma_with_arg(cursor, pragma, arg):
    cursor.execute("PRAGMA {} = '{}'".format(pragma, arg.replace("'", "''")))


def _connect(filename, passphrase, migrate):
    db = makedir_wrapper(sqlite3.connect, filename, sqlite3.OperationalError)

    try:
        cursor = db.cursor()
        cursor.execute('PRAGMA cipher_version')
        if not cursor.fetchone():
            raise DatabaseError('SQLCipher unavailable')

        _execute_pragma_with_arg(cursor, 'key', passphrase)

        # These are the default settings of SQLCipher 4 except
        # for the key derivation algorithm which is the only one
        # supported by SQLCipher 3. That is the best we can do
        # while preserving compatibility with SQLCipher 3.
        cursor.executescript(
            'PRAGMA cipher_page_size = 4096;'
            'PRAGMA kdf_iter = 256000;'
            'PRAGMA cipher_hmac_algorithm = HMAC_SHA1;'
            'PRAGMA cipher_kdf_algorithm = PBKDF2_HMAC_SHA1;'
        )

        try:
            cursor.execute('SELECT COUNT(*) FROM sqlite_master')
        except sqlite3.DatabaseError:
            if migrate:
                db.close()
                from mypass.migration import update_from_legacy_db
                if update_from_legacy_db(filename, passphrase):
                    return _connect(filename, passphrase, migrate=False)
            raise DatabaseError('Wrong passphrase or broken database')

        if cursor.fetchone()[0] == 0:
            cursor.executescript(
                '''CREATE TABLE credentials (
                       id INTEGER NOT NULL,
                       username TEXT NOT NULL,
                       password TEXT NOT NULL,
                       PRIMARY KEY (id, username)
                   ) WITHOUT ROWID;
                   CREATE TABLE contexts (
                       id INTEGER NOT NULL,
                       context TEXT COLLATE NOCASE PRIMARY KEY
                   ) WITHOUT ROWID;
                   CREATE INDEX credentials_id_index ON credentials(id);
                   CREATE INDEX contexts_id_index ON contexts(id);'''
            )
    except:
        db.close()
        raise

    return db


def _get_id_or_create_context(cursor, context):
    cursor.execute('SELECT id FROM contexts '
                   'WHERE context = ?', (context,))
    result = cursor.fetchone()
    if result:
        id = result[0]
    else:
        cursor.execute('''SELECT MAX((SELECT IFNULL(MAX(id), 0) FROM credentials),
                                     (SELECT IFNULL(MAX(id), 0) FROM contexts))''')
        id = cursor.fetchone()[0] + 1
        cursor.execute('INSERT INTO contexts (id, context) '
                       'VALUES (?, ?)', (id, context))
    return id


def _get_id(cursor, context, username):
    cursor.execute('''SELECT id
                        FROM credentials
                  INNER JOIN contexts USING (id)
                       WHERE context = ?
                         AND username = ?''', (context, username))
    result = cursor.fetchone()
    if not result:
        raise CredentialsDoNotExist
    return result[0]


def _clear_orphans(cursor, target_table, updated_table, id):
    cursor.execute('SELECT COUNT(*) FROM {} '
                   'WHERE id = ?'.format(updated_table), (id,))
    if cursor.fetchone()[0] == 0:
        cursor.execute('DELETE FROM {} '
                       'WHERE id = ?'.format(target_table), (id,))


def _rename_or_alias_context(sql, db, old_context, new_context, override):
    with db:
        cursor = db.cursor()

        result = None
        if override:
            cursor.execute('SELECT id FROM contexts '
                           'WHERE context = ?', (new_context,))
            result = cursor.fetchone()

        try:
            cursor.execute(sql.format(' OR REPLACE' if override else ''),
                           (new_context, old_context))
        except sqlite3.IntegrityError:
            raise CredentialsAlreadytExist

        if cursor.rowcount == 0:
            raise CredentialsDoNotExist

        if result:
            _clear_orphans(cursor, 'credentials', 'contexts', result[0])


class Database:
    def __init__(self, filename, passphrase):
        self._db = _connect(filename, passphrase, migrate=True)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def get_credentials(self, context):
        credentials = list(self._db.execute('''SELECT username, password
                                                 FROM credentials
                                           INNER JOIN contexts USING (id)
                                                WHERE context = ?
                                             ORDER BY username''', (context,)))
        if not credentials:
            raise CredentialsDoNotExist
        return credentials

    def get_contexts(self):
        return [ctx for ctx, in self._db.execute('''SELECT context
                                                      FROM contexts
                                                  ORDER BY context''')]

    def store_credentials(self, context, username, password, override=False):
        with self._db:
            cursor = self._db.cursor()
            id = _get_id_or_create_context(cursor, context)

            try:
                cursor.execute(
                    'INSERT{} INTO credentials (id, username, password) '
                    'VALUES (?, ?, ?)'.format(' OR REPLACE' if override else ''),
                    (id, username, password)
                )
            except sqlite3.IntegrityError:
                raise CredentialsAlreadytExist

    def delete_credentials(self, context, username):
        with self._db:
            cursor = self._db.cursor()
            id = _get_id(cursor, context, username)
            cursor.execute('DELETE FROM credentials '
                           'WHERE id = ? AND username = ?''', (id, username))
            _clear_orphans(cursor, 'contexts', 'credentials', id)

    def delete_context(self, context):
        with self._db:
            cursor = self._db.cursor()
            cursor.execute('SELECT id FROM contexts '
                           'WHERE context = ?', (context,))
            result = cursor.fetchone()
            if not result:
                raise CredentialsDoNotExist
            id = result[0]
            cursor.execute('DELETE FROM contexts '
                           'WHERE context = ?', (context,))
            _clear_orphans(cursor, 'credentials', 'contexts', id)

    def rename_credentials(self, old_context, old_username, new_context, new_username, override=False):
        with self._db:
            cursor = self._db.cursor()
            old_id = _get_id(cursor, old_context, old_username)
            new_id = _get_id_or_create_context(cursor, new_context)

            try:
                cursor.execute(
                    '''UPDATE{} credentials
                            SET id = ?,
                                username = ?
                          WHERE id = ?
                            AND username = ?'''.format(' OR REPLACE' if override else ''),
                    (new_id, new_username, old_id, old_username)
                )
            except sqlite3.IntegrityError:
                raise CredentialsAlreadytExist

            _clear_orphans(cursor, 'contexts', 'credentials', old_id)

    def rename_context(self, old_context, new_context, override=False):
        _rename_or_alias_context('''UPDATE{} contexts
                                         SET context = ?
                                       WHERE context = ?''',
                                 self._db, old_context, new_context, override)

    def add_context_alias(self, context, context_alias, override=True):
        _rename_or_alias_context('''INSERT{} INTO contexts (id, context)
                                           SELECT id, ?
                                             FROM contexts
                                            WHERE context = ?''',
                                 self._db, context, context_alias, override)

    def change_passphrase(self, passphrase):
        _execute_pragma_with_arg(self._db, 'rekey', passphrase)

    def close(self):
        self._db.close()



# Copyright (c) 2014-2025 Sebastian Noack
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

import sys
import struct
import json
import re
import urllib.parse

from mypass import CredentialsDoNotExist
from mypass.client import Client, database_exists, ConnectionLost


def parse_request(length_bytes):
    length = struct.unpack('I', length_bytes)[0]
    data = sys.stdin.buffer.read(length)
    return json.loads(data.decode('utf-8'))


def send_response(message):
    data = json.dumps(message).encode('utf-8')
    sys.stdout.buffer.write(struct.pack('I', len(data)))
    sys.stdout.buffer.write(data)
    sys.stdout.buffer.flush()


def get_possible_contexts(url):
    parsed = urllib.parse.urlparse(url)
    parts = parsed.path.split('/')

    while parts:
        yield parsed.netloc + '/'.join(parts)

        tail = parts.pop()
        if tail != '':
            parts.append('')

    if re.search(r'[^0-9.:]', parsed.netloc):
        parts = parsed.netloc.split('.')

        while len(parts) > 2:
            del parts[0]
            yield '.'.join(parts)


class NativeMessagingHost:

    def __init__(self):
        while True:
            length_bytes = sys.stdin.buffer.read(4)
            if not length_bytes:
                break

            send_response(self._process_request(parse_request(length_bytes)))

    def _handle_unlock_database(self, client, request):
        if client.database_locked:
            try:
                client.unlock_database(request['passphrase'])
            except ConnectionLost:
                return {'status': 'failure'}

        return {'status': 'ok'}

    def _handle_lock_database(self, client, request):
        if not client.database_locked:
            client.call('shutdown')

        return {'status': 'ok'}

    def _handle_get_credentials(self, client, request):
        if client.database_locked:
            return {'status': 'database-locked'}

        for context in get_possible_contexts(request['url']):
            try:
                credentials = client.call('get-credentials', context)
            except CredentialsDoNotExist:
                continue

            return {'status': 'ok', 'credentials': [dict(zip(['username', 'password'], token)) for token in credentials]}

        return {'status': 'no-credentials'}

    def _process_request(self, request):
        if not database_exists():
            return {'status': 'database-does-not-exist'}

        with Client() as client:
            return getattr(self, '_handle_' + request['action'].replace('-', '_'))(client, request)



# Copyright (c) 2014-2025 Sebastian Noack
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.

import sys
import os
import pickle
import socket
import select
import signal

from mypass import Error, SOCKET
from mypass.config import get_config
from mypass.db import Database, makedir_wrapper


def make_db_wrapper(method):
    def wrapper(self, *args):
        return getattr(self._db, method)(*args)
    return wrapper


class Daemon:
    def __init__(self, fd_in, fd_out, timeout):
        self._connections = {fd_in: (os.fdopen(fd_in, 'rb', 0, closefd=False),
                                     os.fdopen(fd_out, 'wb', 0, closefd=False))}
        self._poll = select.poll()
        self._poll.register(fd_in, select.POLLIN)
        self._timeout = timeout
        self._db = None
        self._socket = None
        self._shutdown = False

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def _handle_init(self, passphrase, listen):
        self._db = Database(get_config('database', 'path'), passphrase)

        if listen:
            self._socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)

            try:
                os.unlink(SOCKET)
            except FileNotFoundError:
                pass

            self._socket.bind(SOCKET)
            self._socket.listen(5)
            self._poll.register(self._socket, select.POLLIN)

    def _handle_shutdown(self):
        self._shutdown = True

    _handle_get_credentials = make_db_wrapper('get_credentials')
    _handle_get_contexts = make_db_wrapper('get_contexts')
    _handle_store_credentials = make_db_wrapper('store_credentials')
    _handle_delete_credentials = make_db_wrapper('delete_credentials')
    _handle_delete_context = make_db_wrapper('delete_context')
    _handle_rename_credentials = make_db_wrapper('rename_credentials')
    _handle_rename_context = make_db_wrapper('rename_context')
    _handle_change_passphrase = make_db_wrapper('change_passphrase')
    _handle_add_context_alias = make_db_wrapper('add_context_alias')

    def _serve_request(self, reader, writer):
        cmd, args = pickle.load(reader)

        try:
            response = getattr(self, '_handle_' + cmd.replace('-', '_'))(*args)
        except Error as e:
            response = e

        pickle.dump(response, writer)

    def run(self):
        while True:
            events = self._poll.poll(self._timeout)

            for fd, _ in events:
                if self._socket and fd == self._socket.fileno():
                    with self._socket.accept()[0] as conn:
                        self._poll.register(conn, select.POLLIN)
                        self._connections[conn.fileno()] = (
                            conn.makefile('rb', 0),
                            conn.makefile('wb', 0)
                        )
                else:
                    reader, writer = self._connections[fd]
                    try:
                        self._serve_request(reader, writer)
                    except EOFError:
                        self._poll.unregister(fd)
                        del self._connections[fd]
                        reader.close()
                        writer.close()

            if not events or self._shutdown or (not self._socket and not self._connections):
                break

    def close(self):
        for (reader, writer) in self._connections.values():
            reader.close()
            writer.close()
        if self._socket:
            self._socket.close()
        if self._db:
            self._db.close()


if __name__ == '__main__':
    try:
        os.umask(0o077)

        signal.signal(signal.SIGINT, signal.SIG_IGN)
        signal.signal(signal.SIGHUP, signal.SIG_IGN)

        log_fd = makedir_wrapper(lambda fn: os.open(fn, os.O_WRONLY |
                                                        os.O_TRUNC |
                                                        os.O_CREAT, 0o666),
                                 get_config('daemon', 'logfile'))
        os.dup2(log_fd, sys.stderr.fileno())
        os.close(log_fd)

        with Daemon(sys.stdin.fileno(),
                    sys.stdout.fileno(),
                    get_config('daemon', 'timeout') * 60000) as daemon:
            daemon.run()
    except Error as e:
        print(e, file=sys.stderr)
        sys.exit(1)
    except KeyboardInterrupt:
        sys.exit(1)
